from textwrap import TextWrapper
from typing import List

from .utils import get_domain, get_metadatarecord, map_metadatarecord_to_result


class Entity:
    """
    Class to represent every ELG entity
    """

    def __init__(
        self,
        id: int,
        resource_name: str,
        resource_short_name: List[str],
        resource_type: str,
        entity_type: str,
        description: str,
        keywords: List[str],
        detail: str,
        licences: List[str],
        languages: List[str],
        country_of_registration: List[str],
        creation_date: str,
        last_date_updated: str,
        functional_service: bool,
        functions: List[str],
        intended_applications: List[str],
        views: int,
        downloads: int,
        size: int,
        service_execution_count: int,
        status: str,
        under_construction: bool,
        domain: str,
        record: dict = None,
        **kwargs,
    ):
        """Init an Entity object with the entity information"""
        self.id = id
        self.resource_name = resource_name
        self.resource_short_name = resource_short_name
        self.resource_type = resource_type
        self.entity_type = entity_type
        self.description = description
        self.keywords = keywords
        self.detail = detail
        self.licences = licences
        self.languages = languages
        self.country_of_registration = country_of_registration
        self.creation_date = creation_date
        self.last_date_updated = last_date_updated
        self.functional_service = functional_service
        self.functions = functions
        self.intended_applications = intended_applications
        self.views = views
        self.downloads = downloads
        self.size = size
        self.service_execution_count = service_execution_count
        self.status = status
        self.under_construction = under_construction
        self.domain = domain
        self.record = record

    def __str__(self):
        id_wrapper = TextWrapper(initial_indent="Id             ", width=70, subsequent_indent=" " * 15)
        name_wrapper = TextWrapper(initial_indent="Name           ", width=70, subsequent_indent=" " * 15)
        resource_type_wrapper = TextWrapper(initial_indent="Resource type  ", width=70, subsequent_indent=" " * 15)
        entity_type_wrapper = TextWrapper(initial_indent="Entity type    ", width=70, subsequent_indent=" " * 15)
        description_wrapper = TextWrapper(initial_indent="Description    ", width=70, subsequent_indent=" " * 15)
        licences_wrapper = TextWrapper(initial_indent="Licences       ", width=70, subsequent_indent=" " * 15)
        languages_wrapper = TextWrapper(initial_indent="Languages      ", width=70, subsequent_indent=" " * 15)
        status_wrapper = TextWrapper(initial_indent="Status         ", width=70, subsequent_indent=" " * 15)
        return (
            "-" * 70
            + "\n"
            + id_wrapper.fill(str(self.id))
            + "\n"
            + name_wrapper.fill(str(self.resource_name))
            + "\n"
            + resource_type_wrapper.fill(str(self.resource_type))
            + "\n"
            + entity_type_wrapper.fill(str(self.entity_type))
            + "\n"
            + description_wrapper.fill(str(self.description))
            + "\n"
            + licences_wrapper.fill(str(self.licences))
            + "\n"
            + languages_wrapper.fill(str(self.languages))
            + "\n"
            + status_wrapper.fill(str(self.status))
            + "\n"
            + "-" * 70
        )

    def __repr__(self):
        return str(self.__class__)[:-1] + f" [{self.id}] {self.resource_name}>"

    @classmethod
    def from_search_result(cls, result: dict, domain: str = "live"):
        """
        Class method to init an Entity object from a search result.

        Args:
            result (dict): result of the search API.
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".

        Returns:
            elg.Entity: Entity object.
        """
        result["domain"] = get_domain(domain=domain)
        # Add default value to the status parameter as it has been removed from the search API result
        if "status" not in result.keys():
            result["status"] = None
        return cls(**result)

    @classmethod
    def from_id(
        cls,
        id: int,
        domain: str = "live",
        use_cache: bool = True,
        cache_dir: str = "~/.cache/elg",
    ):
        """
        Class method to init an Entity object from its id.

        Args:
            id (int): id of the entity.
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".
            use_cache (bool, optional): True if you want to use cached files. Defaults to True.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files. Defaults to "~/.cache/elg".

        Returns:
            elg.Entity: Entity object.
        """
        domain = get_domain(domain=domain)
        record = get_metadatarecord(id=id, domain=domain, use_cache=use_cache, cache_dir=cache_dir)
        result = map_metadatarecord_to_result(id, record)
        result["domain"] = domain
        return cls(**result)



#!/usr/bin/env python3

from collections import defaultdict

# Data below is from https://www.loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt
ISO639_DATA = """\
aar||aa|Afar|afar
abk||ab|Abkhazian|abkhaze
ace|||Achinese|aceh
ach|||Acoli|acoli
ada|||Adangme|adangme
ady|||Adyghe; Adygei|adyghé
afa|||Afro-Asiatic languages|afro-asiatiques, langues
afh|||Afrihili|afrihili
afr||af|Afrikaans|afrikaans
ain|||Ainu|aïnou
aka||ak|Akan|akan
akk|||Akkadian|akkadien
alb|sqi|sq|Albanian|albanais
ale|||Aleut|aléoute
alg|||Algonquian languages|algonquines, langues
alt|||Southern Altai|altai du Sud
amh||am|Amharic|amharique
ang|||English, Old (ca.450-1100)|anglo-saxon (ca.450-1100)
anp|||Angika|angika
apa|||Apache languages|apaches, langues
ara||ar|Arabic|arabe
arc|||Official Aramaic (700-300 BCE); Imperial Aramaic (700-300 BCE)|araméen d'empire (700-300 BCE)
arg||an|Aragonese|aragonais
arm|hye|hy|Armenian|arménien
arn|||Mapudungun; Mapuche|mapudungun; mapuche; mapuce
arp|||Arapaho|arapaho
art|||Artificial languages|artificielles, langues
arw|||Arawak|arawak
asm||as|Assamese|assamais
ast|||Asturian; Bable; Leonese; Asturleonese|asturien; bable; léonais; asturoléonais
ath|||Athapascan languages|athapascanes, langues
aus|||Australian languages|australiennes, langues
ava||av|Avaric|avar
ave||ae|Avestan|avestique
awa|||Awadhi|awadhi
aym||ay|Aymara|aymara
aze||az|Azerbaijani|azéri
bad|||Banda languages|banda, langues
bai|||Bamileke languages|bamiléké, langues
bak||ba|Bashkir|bachkir
bal|||Baluchi|baloutchi
bam||bm|Bambara|bambara
ban|||Balinese|balinais
baq|eus|eu|Basque|basque
bas|||Basa|basa
bat|||Baltic languages|baltes, langues
bej|||Beja; Bedawiyet|bedja
bel||be|Belarusian|biélorusse
bem|||Bemba|bemba
ben||bn|Bengali|bengali
ber|||Berber languages|berbères, langues
bho|||Bhojpuri|bhojpuri
bih||bh|Bihari languages|langues biharis
bik|||Bikol|bikol
bin|||Bini; Edo|bini; edo
bis||bi|Bislama|bichlamar
bla|||Siksika|blackfoot
bnt|||Bantu languages|bantou, langues
bos||bs|Bosnian|bosniaque
bra|||Braj|braj
bre||br|Breton|breton
btk|||Batak languages|batak, langues
bua|||Buriat|bouriate
bug|||Buginese|bugi
bul||bg|Bulgarian|bulgare
bur|mya|my|Burmese|birman
byn|||Blin; Bilin|blin; bilen
cad|||Caddo|caddo
cai|||Central American Indian languages|amérindiennes de L'Amérique centrale, langues
car|||Galibi Carib|karib; galibi; carib
cat||ca|Catalan; Valencian|catalan; valencien
cau|||Caucasian languages|caucasiennes, langues
ceb|||Cebuano|cebuano
cel|||Celtic languages|celtiques, langues; celtes, langues
cha||ch|Chamorro|chamorro
chb|||Chibcha|chibcha
che||ce|Chechen|tchétchène
chg|||Chagatai|djaghataï
chi|zho|zh|Chinese|chinois
chk|||Chuukese|chuuk
chm|||Mari|mari
chn|||Chinook jargon|chinook, jargon
cho|||Choctaw|choctaw
chp|||Chipewyan; Dene Suline|chipewyan
chr|||Cherokee|cherokee
chu||cu|Church Slavic; Old Slavonic; Church Slavonic; Old Bulgarian; Old Church Slavonic|slavon d'église; vieux slave; slavon liturgique; vieux bulgare
chv||cv|Chuvash|tchouvache
chy|||Cheyenne|cheyenne
cmc|||Chamic languages|chames, langues
cnr|||Montenegrin|monténégrin
cop|||Coptic|copte
cor||kw|Cornish|cornique
cos||co|Corsican|corse
cpe|||Creoles and pidgins, English based|créoles et pidgins basés sur l'anglais
cpf|||Creoles and pidgins, French-based|créoles et pidgins basés sur le français
cpp|||Creoles and pidgins, Portuguese-based|créoles et pidgins basés sur le portugais
cre||cr|Cree|cree
crh|||Crimean Tatar; Crimean Turkish|tatar de Crimé
crp|||Creoles and pidgins|créoles et pidgins
csb|||Kashubian|kachoube
cus|||Cushitic languages|couchitiques, langues
cze|ces|cs|Czech|tchèque
dak|||Dakota|dakota
dan||da|Danish|danois
dar|||Dargwa|dargwa
day|||Land Dayak languages|dayak, langues
del|||Delaware|delaware
den|||Slave (Athapascan)|esclave (athapascan)
dgr|||Dogrib|dogrib
din|||Dinka|dinka
div||dv|Divehi; Dhivehi; Maldivian|maldivien
doi|||Dogri|dogri
dra|||Dravidian languages|dravidiennes, langues
dsb|||Lower Sorbian|bas-sorabe
dua|||Duala|douala
dum|||Dutch, Middle (ca.1050-1350)|néerlandais moyen (ca. 1050-1350)
dut|nld|nl|Dutch; Flemish|néerlandais; flamand
dyu|||Dyula|dioula
dzo||dz|Dzongkha|dzongkha
efi|||Efik|efik
egy|||Egyptian (Ancient)|égyptien
eka|||Ekajuk|ekajuk
elx|||Elamite|élamite
eng||en|English|anglais
enm|||English, Middle (1100-1500)|anglais moyen (1100-1500)
epo||eo|Esperanto|espéranto
est||et|Estonian|estonien
ewe||ee|Ewe|éwé
ewo|||Ewondo|éwondo
fan|||Fang|fang
fao||fo|Faroese|féroïen
fat|||Fanti|fanti
fij||fj|Fijian|fidjien
fil|||Filipino; Pilipino|filipino; pilipino
fin||fi|Finnish|finnois
fiu|||Finno-Ugrian languages|finno-ougriennes, langues
fon|||Fon|fon
fre|fra|fr|French|français
frm|||French, Middle (ca.1400-1600)|français moyen (1400-1600)
fro|||French, Old (842-ca.1400)|français ancien (842-ca.1400)
frr|||Northern Frisian|frison septentrional
frs|||Eastern Frisian|frison oriental
fry||fy|Western Frisian|frison occidental
ful||ff|Fulah|peul
fur|||Friulian|frioulan
gaa|||Ga|ga
gay|||Gayo|gayo
gba|||Gbaya|gbaya
gem|||Germanic languages|germaniques, langues
geo|kat|ka|Georgian|géorgien
ger|deu|de|German|allemand
gez|||Geez|guèze
gil|||Gilbertese|kiribati
gla||gd|Gaelic; Scottish Gaelic|gaélique; gaélique écossais
gle||ga|Irish|irlandais
glg||gl|Galician|galicien
glv||gv|Manx|manx; mannois
gmh|||German, Middle High (ca.1050-1500)|allemand, moyen haut (ca. 1050-1500)
goh|||German, Old High (ca.750-1050)|allemand, vieux haut (ca. 750-1050)
gon|||Gondi|gond
gor|||Gorontalo|gorontalo
got|||Gothic|gothique
grb|||Grebo|grebo
grc|||Greek, Ancient (to 1453)|grec ancien (jusqu'à 1453)
gre|ell|el|Greek, Modern (1453-)|grec moderne (après 1453)
grn||gn|Guarani|guarani
gsw|||Swiss German; Alemannic; Alsatian|suisse alémanique; alémanique; alsacien
guj||gu|Gujarati|goudjrati
gwi|||Gwich'in|gwich'in
hai|||Haida|haida
hat||ht|Haitian; Haitian Creole|haïtien; créole haïtien
hau||ha|Hausa|haoussa
haw|||Hawaiian|hawaïen
heb||he|Hebrew|hébreu
her||hz|Herero|herero
hil|||Hiligaynon|hiligaynon
him|||Himachali languages; Western Pahari languages|langues himachalis; langues paharis occidentales
hin||hi|Hindi|hindi
hit|||Hittite|hittite
hmn|||Hmong; Mong|hmong
hmo||ho|Hiri Motu|hiri motu
hrv||hr|Croatian|croate
hsb|||Upper Sorbian|haut-sorabe
hun||hu|Hungarian|hongrois
hup|||Hupa|hupa
iba|||Iban|iban
ibo||ig|Igbo|igbo
ice|isl|is|Icelandic|islandais
ido||io|Ido|ido
iii||ii|Sichuan Yi; Nuosu|yi de Sichuan
ijo|||Ijo languages|ijo, langues
iku||iu|Inuktitut|inuktitut
ile||ie|Interlingue; Occidental|interlingue
ilo|||Iloko|ilocano
ina||ia|Interlingua (International Auxiliary Language Association)|interlingua (langue auxiliaire internationale)
inc|||Indic languages|indo-aryennes, langues
ind||id|Indonesian|indonésien
ine|||Indo-European languages|indo-européennes, langues
inh|||Ingush|ingouche
ipk||ik|Inupiaq|inupiaq
ira|||Iranian languages|iraniennes, langues
iro|||Iroquoian languages|iroquoises, langues
ita||it|Italian|italien
jav||jv|Javanese|javanais
jbo|||Lojban|lojban
jpn||ja|Japanese|japonais
jpr|||Judeo-Persian|judéo-persan
jrb|||Judeo-Arabic|judéo-arabe
kaa|||Kara-Kalpak|karakalpak
kab|||Kabyle|kabyle
kac|||Kachin; Jingpho|kachin; jingpho
kal||kl|Kalaallisut; Greenlandic|groenlandais
kam|||Kamba|kamba
kan||kn|Kannada|kannada
kar|||Karen languages|karen, langues
kas||ks|Kashmiri|kashmiri
kau||kr|Kanuri|kanouri
kaw|||Kawi|kawi
kaz||kk|Kazakh|kazakh
kbd|||Kabardian|kabardien
kha|||Khasi|khasi
khi|||Khoisan languages|khoïsan, langues
khm||km|Central Khmer|khmer central
kho|||Khotanese; Sakan|khotanais; sakan
kik||ki|Kikuyu; Gikuyu|kikuyu
kin||rw|Kinyarwanda|rwanda
kir||ky|Kirghiz; Kyrgyz|kirghiz
kmb|||Kimbundu|kimbundu
kok|||Konkani|konkani
kom||kv|Komi|kom
kon||kg|Kongo|kongo
kor||ko|Korean|coréen
kos|||Kosraean|kosrae
kpe|||Kpelle|kpellé
krc|||Karachay-Balkar|karatchai balkar
krl|||Karelian|carélien
kro|||Kru languages|krou, langues
kru|||Kurukh|kurukh
kua||kj|Kuanyama; Kwanyama|kuanyama; kwanyama
kum|||Kumyk|koumyk
kur||ku|Kurdish|kurde
kut|||Kutenai|kutenai
lad|||Ladino|judéo-espagnol
lah|||Lahnda|lahnda
lam|||Lamba|lamba
lao||lo|Lao|lao
lat||la|Latin|latin
lav||lv|Latvian|letton
lez|||Lezghian|lezghien
lim||li|Limburgan; Limburger; Limburgish|limbourgeois
lin||ln|Lingala|lingala
lit||lt|Lithuanian|lituanien
lol|||Mongo|mongo
loz|||Lozi|lozi
ltz||lb|Luxembourgish; Letzeburgesch|luxembourgeois
lua|||Luba-Lulua|luba-lulua
lub||lu|Luba-Katanga|luba-katanga
lug||lg|Ganda|ganda
lui|||Luiseno|luiseno
lun|||Lunda|lunda
luo|||Luo (Kenya and Tanzania)|luo (Kenya et Tanzanie)
lus|||Lushai|lushai
mac|mkd|mk|Macedonian|macédonien
mad|||Madurese|madourais
mag|||Magahi|magahi
mah||mh|Marshallese|marshall
mai|||Maithili|maithili
mak|||Makasar|makassar
mal||ml|Malayalam|malayalam
man|||Mandingo|mandingue
mao|mri|mi|Maori|maori
map|||Austronesian languages|austronésiennes, langues
mar||mr|Marathi|marathe
mas|||Masai|massaï
may|msa|ms|Malay|malais
mdf|||Moksha|moksa
mdr|||Mandar|mandar
men|||Mende|mendé
mga|||Irish, Middle (900-1200)|irlandais moyen (900-1200)
mic|||Mi'kmaq; Micmac|mi'kmaq; micmac
min|||Minangkabau|minangkabau
mis|||Uncoded languages|langues non codées
mkh|||Mon-Khmer languages|môn-khmer, langues
mlg||mg|Malagasy|malgache
mlt||mt|Maltese|maltais
mnc|||Manchu|mandchou
mni|||Manipuri|manipuri
mno|||Manobo languages|manobo, langues
moh|||Mohawk|mohawk
mon||mn|Mongolian|mongol
mos|||Mossi|moré
mul|||Multiple languages|multilingue
mun|||Munda languages|mounda, langues
mus|||Creek|muskogee
mwl|||Mirandese|mirandais
mwr|||Marwari|marvari
myn|||Mayan languages|maya, langues
myv|||Erzya|erza
nah|||Nahuatl languages|nahuatl, langues
nai|||North American Indian languages|nord-amérindiennes, langues
nap|||Neapolitan|napolitain
nau||na|Nauru|nauruan
nav||nv|Navajo; Navaho|navaho
nbl||nr|Ndebele, South; South Ndebele|ndébélé du Sud
nde||nd|Ndebele, North; North Ndebele|ndébélé du Nord
ndo||ng|Ndonga|ndonga
nds|||Low German; Low Saxon; German, Low; Saxon, Low|bas allemand; bas saxon; allemand, bas; saxon, bas
nep||ne|Nepali|népalais
new|||Nepal Bhasa; Newari|nepal bhasa; newari
nia|||Nias|nias
nic|||Niger-Kordofanian languages|nigéro-kordofaniennes, langues
niu|||Niuean|niué
nno||nn|Norwegian Nynorsk; Nynorsk, Norwegian|norvégien nynorsk; nynorsk, norvégien
nob||nb|Bokmål, Norwegian; Norwegian Bokmål|norvégien bokmål
nog|||Nogai|nogaï; nogay
non|||Norse, Old|norrois, vieux
nor||no|Norwegian|norvégien
nqo|||N'Ko|n'ko
nso|||Pedi; Sepedi; Northern Sotho|pedi; sepedi; sotho du Nord
nub|||Nubian languages|nubiennes, langues
nwc|||Classical Newari; Old Newari; Classical Nepal Bhasa|newari classique
nya||ny|Chichewa; Chewa; Nyanja|chichewa; chewa; nyanja
nym|||Nyamwezi|nyamwezi
nyn|||Nyankole|nyankolé
nyo|||Nyoro|nyoro
nzi|||Nzima|nzema
oci||oc|Occitan (post 1500)|occitan (après 1500)
oji||oj|Ojibwa|ojibwa
ori||or|Oriya|oriya
orm||om|Oromo|galla
osa|||Osage|osage
oss||os|Ossetian; Ossetic|ossète
ota|||Turkish, Ottoman (1500-1928)|turc ottoman (1500-1928)
oto|||Otomian languages|otomi, langues
paa|||Papuan languages|papoues, langues
pag|||Pangasinan|pangasinan
pal|||Pahlavi|pahlavi
pam|||Pampanga; Kapampangan|pampangan
pan||pa|Panjabi; Punjabi|pendjabi
pap|||Papiamento|papiamento
pau|||Palauan|palau
peo|||Persian, Old (ca.600-400 B.C.)|perse, vieux (ca. 600-400 av. J.-C.)
per|fas|fa|Persian|persan
phi|||Philippine languages|philippines, langues
phn|||Phoenician|phénicien
pli||pi|Pali|pali
pol||pl|Polish|polonais
pon|||Pohnpeian|pohnpei
por||pt|Portuguese|portugais
pra|||Prakrit languages|prâkrit, langues
pro|||Provençal, Old (to 1500); Occitan, Old (to 1500)|provençal ancien (jusqu'à 1500); occitan ancien (jusqu'à 1500)
pus||ps|Pushto; Pashto|pachto
qaa-qtz|||Reserved for local use|réservée à l'usage local
que||qu|Quechua|quechua
raj|||Rajasthani|rajasthani
rap|||Rapanui|rapanui
rar|||Rarotongan; Cook Islands Maori|rarotonga; maori des îles Cook
roa|||Romance languages|romanes, langues
roh||rm|Romansh|romanche
rom|||Romany|tsigane
rum|ron|ro|Romanian; Moldavian; Moldovan|roumain; moldave
run||rn|Rundi|rundi
rup|||Aromanian; Arumanian; Macedo-Romanian|aroumain; macédo-roumain
rus||ru|Russian|russe
sad|||Sandawe|sandawe
sag||sg|Sango|sango
sah|||Yakut|iakoute
sai|||South American Indian languages|sud-amérindiennes, langues
sal|||Salishan languages|salishennes, langues
sam|||Samaritan Aramaic|samaritain
san||sa|Sanskrit|sanskrit
sas|||Sasak|sasak
sat|||Santali|santal
scn|||Sicilian|sicilien
sco|||Scots|écossais
sel|||Selkup|selkoupe
sem|||Semitic languages|sémitiques, langues
sga|||Irish, Old (to 900)|irlandais ancien (jusqu'à 900)
sgn|||Sign Languages|langues des signes
shn|||Shan|chan
sid|||Sidamo|sidamo
sin||si|Sinhala; Sinhalese|singhalais
sio|||Siouan languages|sioux, langues
sit|||Sino-Tibetan languages|sino-tibétaines, langues
sla|||Slavic languages|slaves, langues
slo|slk|sk|Slovak|slovaque
slv||sl|Slovenian|slovène
sma|||Southern Sami|sami du Sud
sme||se|Northern Sami|sami du Nord
smi|||Sami languages|sames, langues
smj|||Lule Sami|sami de Lule
smn|||Inari Sami|sami d'Inari
smo||sm|Samoan|samoan
sms|||Skolt Sami|sami skolt
sna||sn|Shona|shona
snd||sd|Sindhi|sindhi
snk|||Soninke|soninké
sog|||Sogdian|sogdien
som||so|Somali|somali
son|||Songhai languages|songhai, langues
sot||st|Sotho, Southern|sotho du Sud
spa||es|Spanish; Castilian|espagnol; castillan
srd||sc|Sardinian|sarde
srn|||Sranan Tongo|sranan tongo
srp||sr|Serbian|serbe
srr|||Serer|sérère
ssa|||Nilo-Saharan languages|nilo-sahariennes, langues
ssw||ss|Swati|swati
suk|||Sukuma|sukuma
sun||su|Sundanese|soundanais
sus|||Susu|soussou
sux|||Sumerian|sumérien
swa||sw|Swahili|swahili
swe||sv|Swedish|suédois
syc|||Classical Syriac|syriaque classique
syr|||Syriac|syriaque
tah||ty|Tahitian|tahitien
tai|||Tai languages|tai, langues
tam||ta|Tamil|tamoul
tat||tt|Tatar|tatar
tel||te|Telugu|télougou
tem|||Timne|temne
ter|||Tereno|tereno
tet|||Tetum|tetum
tgk||tg|Tajik|tadjik
tgl||tl|Tagalog|tagalog
tha||th|Thai|thaï
tib|bod|bo|Tibetan|tibétain
tig|||Tigre|tigré
tir||ti|Tigrinya|tigrigna
tiv|||Tiv|tiv
tkl|||Tokelau|tokelau
tlh|||Klingon; tlhIngan-Hol|klingon
tli|||Tlingit|tlingit
tmh|||Tamashek|tamacheq
tog|||Tonga (Nyasa)|tonga (Nyasa)
ton||to|Tonga (Tonga Islands)|tongan (Îles Tonga)
tpi|||Tok Pisin|tok pisin
tsi|||Tsimshian|tsimshian
tsn||tn|Tswana|tswana
tso||ts|Tsonga|tsonga
tuk||tk|Turkmen|turkmène
tum|||Tumbuka|tumbuka
tup|||Tupi languages|tupi, langues
tur||tr|Turkish|turc
tut|||Altaic languages|altaïques, langues
tvl|||Tuvalu|tuvalu
twi||tw|Twi|twi
tyv|||Tuvinian|touva
udm|||Udmurt|oudmourte
uga|||Ugaritic|ougaritique
uig||ug|Uighur; Uyghur|ouïgour
ukr||uk|Ukrainian|ukrainien
umb|||Umbundu|umbundu
und|||Undetermined|indéterminée
urd||ur|Urdu|ourdou
uzb||uz|Uzbek|ouszbek
vai|||Vai|vaï
ven||ve|Venda|venda
vie||vi|Vietnamese|vietnamien
vol||vo|Volapük|volapük
vot|||Votic|vote
wak|||Wakashan languages|wakashanes, langues
wal|||Wolaitta; Wolaytta|wolaitta; wolaytta
war|||Waray|waray
was|||Washo|washo
wel|cym|cy|Welsh|gallois
wen|||Sorbian languages|sorabes, langues
wln||wa|Walloon|wallon
wol||wo|Wolof|wolof
xal|||Kalmyk; Oirat|kalmouk; oïrat
xho||xh|Xhosa|xhosa
yao|||Yao|yao
yap|||Yapese|yapois
yid||yi|Yiddish|yiddish
yor||yo|Yoruba|yoruba
ypk|||Yupik languages|yupik, langues
zap|||Zapotec|zapotèque
zbl|||Blissymbols; Blissymbolics; Bliss|symboles Bliss; Bliss
zen|||Zenaga|zenaga
zgh|||Standard Moroccan Tamazight|amazighe standard marocain
zha||za|Zhuang; Chuang|zhuang; chuang
znd|||Zande languages|zandé, langues
zul||zu|Zulu|zoulou
zun|||Zuni|zuni
zxx|||No linguistic content; Not applicable|pas de contenu linguistique; non applicable
zza|||Zaza; Dimili; Dimli; Kirdki; Kirmanjki; Zazaki|zaza; dimili; dimli; kirdki; kirmanjki; zazaki
"""

EnglishDescription = {}  # English name(s) of languages (list)
FrenchDescription = {}  # French names(s) of languages (list)
ISO639Codes = defaultdict(list)
for line in ISO639_DATA.strip().split("\n"):
    e3, n3, e2, enames, fnames = line.strip().split("|")
    codes = [x for x in [e3, n3, e2] if len(x)]
    enames = [e.strip() for e in enames.split(";")]
    fnames = [f.strip() for f in fnames.split(";")]
    for code in codes:
        EnglishDescription[code] = enames
        FrenchDescription[code] = fnames
        pass
    for n in enames:
        ISO639Codes[n].extend(codes)
        pass
    pass


class ISO639:
    def __get__(self, key):
        if len(key) >= 2 and key[0].isupper() and key[1].islower():
            return ISO639Codes[key]
        return EnglishDescription[key]

    def LanguageName(self, code):
        return EnglishDescription.get(code, [code])[0]

    def LanguageCodes(self, name):
        return ISO639Codes.get(name, name)

    pass



__version__ = "0.4.7"

import importlib.util

_docker_available = importlib.util.find_spec("docker") is not None
_docker_available = importlib.util.find_spec("flask") is not None
_docker_available = importlib.util.find_spec("flask_json") is not None
_requests_toolbelt_available = importlib.util.find_spec("requests_toolbelt") is not None


from .authentication import Authentication
from .benchmark import Benchmark
from .catalog import Catalog
from .corpus import Corpus
from .entity import Entity
from .pipeline import Pipeline
from .provider import Provider
from .service import Service

if _docker_available and _docker_available and _docker_available and _requests_toolbelt_available:
    from .flask_service import FlaskService



from base64 import b64decode
from pathlib import Path
from typing import Dict, List

from pydantic import validator

from .. import Annotation, ResponseObject


class AudioResponse(ResponseObject):
    """
    Response representing audio data with optional standoff annotations (e.g. text-to-speech results)

    Attributes
    ----------
    type (string, required): type of response
    warnings (List[StatusMessage], optional): messages describing any warnings on responses
    content (str, required): base64 encoded audio for short audio snippets
    format (str, required): specifies audio format used: either "LINEAR16" or "MP3"
    features (dict, optional): arbitrary json metadata about content
    annotations (Dict[str, List[Annotation]], optional): optional annotations on request
    """

    type: str = "audio"
    content: str
    format: str
    features: dict = None
    annotations: Dict[str, List[Annotation]] = None

    @validator("format")
    def format_must_be_specific(cls, v):
        """
        validator: ensures the format of the audio response is either "LINEAR16" or "MP3"
        """
        acceptable_formats = ["LINEAR16", "MP3"]
        if v.lower() not in acceptable_formats:
            raise ValueError("The format given is not supported")
        return v.lower()

    def to_file(self, filename):
        filename = Path(filename)
        with open(filename, "wb") as f:
            f.write(b64decode(self.content))

    def auto_content(self):
        if self.warnings is not None:
            return self.dict()
        if self.features is not None:
            return self.dict()
        if self.annotations is not None:
            return self.dict()
        return {
            "content": self.dict()["content"],
            "format": self.dict()["format"],
        }



import hashlib
import json
import pickle
from os.path import expanduser
from pathlib import Path
from typing import Dict, List

import requests

from ..model import Response
from .errors import (MissingInformationException, catch_requests_error,
                     ensure_response_ok)
from .ISO639 import ISO639 as iso639

ISO639 = iso639()

LIVE_DOMAIN = "https://live.european-language-grid.eu"
DEV_DOMAIN = "https://dev.european-language-grid.eu"

API_URL = "{}/catalogue_backend/api/registry/"
LICENCE_URL = "{domain}/catalogue_backend/static/project/licences/{licence}.pdf"
CORPUS_DOWNLOAD_URL = "{domain}/catalogue_backend/api/management/download/{id}/"
XML_UPLOAD_URL = "{}/catalogue_backend/api/registry/xmlmetadatarecord/"
S3_UPLOAD_URL = "{}/s3/"


def get_domain(domain):
    if domain == "live":
        return LIVE_DOMAIN
    elif domain == "dev":
        return DEV_DOMAIN
    elif isinstance(domain, str):
        return domain
    else:
        raise ValueError(
            "Domain argument must be a string. Choose 'live' to use the live cluster, 'dev' to use the dev cluster "
            "or 'custom.domain.name' to use a local cluster."
        )


@catch_requests_error
def get_metadatarecord(id: int, domain: str, use_cache: bool, cache_dir: str):
    if use_cache is True and cache_dir is None:
        raise ValueError("To use the cache you have to specify the cache_dir parameters.")

    elif cache_dir is None:  # do not save the response result into the cache_dir
        response = requests.get(API_URL.format(domain) + f"metadatarecord/{id}")
        ensure_response_ok(response)

    else:
        cache_dir = expanduser(cache_dir)
        cache_dir = Path(cache_dir) / Path("metadatarecords")
        cache_dir.mkdir(parents=True, exist_ok=True)
        request_hash = hashlib.sha256(bytes(f"get_metadatarecord - {id} - {domain}", "utf-8")).hexdigest()
        cache_filename = cache_dir / Path(request_hash)

        if use_cache is True and cache_filename.is_file():
            with open(cache_filename, "rb") as cache_file:
                response = pickle.load(cache_file)
        else:
            response = requests.get(API_URL.format(domain) + f"metadatarecord/{id}")
            ensure_response_ok(response)
            with open(cache_filename, "wb") as cache_file:
                pickle.dump(response, cache_file)

    record = response.json()
    if record is None:
        raise ValueError(f"Metadata record for id {id} is null")
    return record


def map_metadatarecord_to_result(id, record):
    described_entity = get_information(id, record, "described_entity")
    entity_type = get_information(id, described_entity, "entity_type")

    if entity_type == "LanguageResource":
        resource_name = get_en_value(get_information(id, described_entity, "resource_name"))
        resource_short_name = get_information(id, described_entity, "resource_short_name", none=True, values=True)
        resource_type = get_information(id, described_entity, ["lr_subclass", "lr_type"])
        description = get_en_value(get_information(id, described_entity, "description"))
        keywords = list(map(get_en_value, get_information(id, described_entity, "keyword")))
        detail = None
        licences = [
            name
            for terms in list(
                map(
                    lambda x: list(
                        map(
                            lambda x: get_en_value(get_information(id, x, "licence_terms_name")),
                            get_information(id, x, "licence_terms"),
                        )
                    ),
                    get_information(
                        id,
                        described_entity,
                        [
                            "lr_subclass",
                            "software_distribution" if resource_type == "ToolService" else "dataset_distribution",
                        ],
                    ),
                )
            )
            for name in terms
        ]
        if resource_type == "ToolService":
            # input_content_resource
            input_content_resource = get_information(id, described_entity, ["lr_subclass", "input_content_resource"])
            # output_resource
            output_resource = get_information(id, described_entity, ["lr_subclass", "output_resource"])
            languages = []
            for resource in input_content_resource + output_resource:
                languages += [
                    ISO639.LanguageName(l["language_tag"]) for l in get_information(id, resource, "language")
                ]
        elif resource_type == "Corpus":
            corpus_media_part = get_information(id, described_entity, ["lr_subclass", "corpus_media_part"])
            languages = []
            for media in corpus_media_part:
                languages += [ISO639.LanguageName(l["language_tag"]) for l in get_information(id, media, "language")]
        elif resource_type == "LanguageDescription":
            language_description_media_part = get_information(
                id, described_entity, ["lr_subclass", "language_description_media_part"]
            )
            languages = []
            for media in language_description_media_part:
                languages += [ISO639.LanguageName(l["language_tag"]) for l in get_information(id, media, "language")]
        elif resource_type == "LexicalConceptualResource":
            lexical_conceptual_resource_media_part = get_information(
                id, described_entity, ["lr_subclass", "lexical_conceptual_resource_media_part"]
            )
            languages = []
            for media in lexical_conceptual_resource_media_part:
                languages += [ISO639.LanguageName(l["language_tag"]) for l in get_information(id, media, "language")]
        else:
            raise ValueError(f"Resource type: {resource_type} not implemented yet.")
        country_of_registration = None
        creation_date = get_information(id, record, "metadata_creation_date")
        last_date_updated = get_information(id, record, "metadata_last_date_updated")
        functional_service = get_information(id, record, ["management_object", "functional_service"])
        functions = None
        intended_applications = None
        views = None
        downloads = None
        size = get_information(id, record, ["management_object", "size"])
        service_execution_count = None
        status = get_information(id, record, ["management_object", "status"])
        under_construction = get_information(id, record, ["management_object", "under_construction"])

    elif entity_type == "Organization":
        resource_name = get_en_value(get_information(id, described_entity, "organization_name"))
        resource_short_name = get_information(id, described_entity, "organization_short_name", none=True, values=True)
        resource_type = None
        description = ""
        keywords = []
        detail = None
        licences = []
        languages = []
        country_of_registration = None
        creation_date = get_information(id, record, "metadata_creation_date")
        last_date_updated = get_information(id, record, "metadata_last_date_updated")
        functional_service = get_information(id, record, ["management_object", "functional_service"])
        functions = None
        intended_applications = None
        views = None
        downloads = None
        size = get_information(id, record, ["management_object", "size"])
        service_execution_count = None
        status = get_information(id, record, ["management_object", "status"])
        under_construction = get_information(id, record, ["management_object", "under_construction"])

    elif entity_type == "Project":
        resource_name = get_en_value(get_information(id, described_entity, "project_name"))
        resource_short_name = get_information(id, described_entity, "project_short_name", none=True, values=True)
        resource_type = None
        description = get_en_value(get_information(id, described_entity, "project_summary"))
        try:
            keywords = list(map(get_en_value, get_information(id, described_entity, "keyword")))
        except:
            keywords = []
        detail = None
        licences = []
        languages = []
        country_of_registration = None
        creation_date = get_information(id, record, "metadata_creation_date")
        last_date_updated = get_information(id, record, "metadata_last_date_updated")
        functional_service = get_information(id, record, ["management_object", "functional_service"])
        functions = None
        intended_applications = None
        views = None
        downloads = None
        size = get_information(id, record, ["management_object", "size"])
        service_execution_count = None
        status = get_information(id, record, ["management_object", "status"])
        under_construction = get_information(id, record, ["management_object", "under_construction"])

    else:
        raise ValueError(f"Entity type: {entity_type} not implemented yet.")

    return {
        "id": id,
        "resource_name": resource_name,
        "resource_short_name": resource_short_name,
        "resource_type": resource_type,
        "entity_type": entity_type,
        "description": description,
        "keywords": keywords,
        "detail": detail,
        "licences": licences,
        "languages": languages,
        "country_of_registration": country_of_registration,
        "creation_date": creation_date,
        "last_date_updated": last_date_updated,
        "functional_service": functional_service,
        "functions": functions,
        "intended_applications": intended_applications,
        "views": views,
        "downloads": downloads,
        "size": size,
        "service_execution_count": service_execution_count,
        "status": status,
        "under_construction": under_construction,
        "record": record,
    }


def get_information(id, obj, infos, none=False, values=False):
    if isinstance(infos, str):
        infos = [infos]
    for info in infos:
        obj = obj.get(info)
        if obj is None:
            if info == infos[-1] and none:
                continue
            raise MissingInformationException(id, info)
    if values:
        if isinstance(obj, dict):
            return list(obj.values())
        return obj
    return obj


def get_en_value(d):
    if "en" in d.keys():
        return d["en"]
    values = list(d.values())
    assert len(values) >= 1
    return values[0]


def get_argument_from_json(filename, args):
    with open(filename, "r") as f:
        tokens = json.load(f)
    if isinstance(args, str):
        return tokens[args]
    return [tokens[arg] for arg in args]



from .base import (Annotation, Failure, Progress, Request, Response,
                   ResponseObject, StatusMessage)
from .request import AudioRequest, StructuredTextRequest, TextRequest
from .response import (AnnotationsResponse, AudioResponse, ClassesResponse,
                       ClassificationResponse, TextsResponse,
                       TextsResponseObject, get_response)



from pydantic import BaseModel

from .StatusMessage import StatusMessage


class Progress(BaseModel):
    """
    Details of an in progress task

    Attributes
    ----------
    percent (float, required): completion percentage
    message (StatusMessage, optional): message describing progress report
    """

    percent: float
    message: StatusMessage = None



import sys
from argparse import ArgumentParser

from loguru import logger

from . import BaseELGCommand


def run_command_factory(args):
    return RunCommand(
        id=args.id,
        auth_file=args.auth_file,
        scope=args.scope,
        domain=args.domain,
        not_use_cache=args.not_use_cache,
        cache_dir=args.cache_dir,
        request_input=args.request_input,
        request_type=args.request_type,
        sync_mode=args.sync_mode,
        timeout=args.timeout,
        not_check_file=args.not_check_file,
    )


class RunCommand(BaseELGCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        run_parser = parser.add_parser("run", description="Run an ELG service")
        run_parser.add_argument("--id", type=int, default=None, required=True, help="ID of the service to run")
        run_parser.add_argument(
            "--auth_file", type=str, default=None, required=False, help="path to the authentification file"
        )
        run_parser.add_argument(
            "--scope",
            type=str,
            default="openid",
            required=False,
            help="Oauth2 scope to use. 'openid' or 'offline_access'",
            choices=["openid", "offline_access"],
        )
        run_parser.add_argument(
            "--domain",
            type=str,
            default="live",
            required=False,
            help="ELG domain you want to use",
        )
        run_parser.add_argument(
            "--not_use_cache",
            default=False,
            action="store_true",
            required=False,
            help="use to not use cache",
        )
        run_parser.add_argument(
            "--cache_dir", type=str, default="~/.cache/elg", required=False, help="path to the cache directory"
        )
        run_parser.add_argument(
            "--request_input",
            type=str,
            default=None,
            required=True,
            help="can be the text to process directly, the name of the file to process, a list of texts, or directly a Request object.",
        )
        run_parser.add_argument(
            "--request_type",
            type=str,
            default="text",
            required=False,
            help="Can be 'text', 'structuredText', or 'audio'",
        )
        run_parser.add_argument(
            "--sync_mode",
            default=False,
            action="store_true",
            required=False,
            help="use to call the service in sync mode",
        )
        run_parser.add_argument(
            "--timeout", type=int, default=None, required=False, help="number of seconds before timeout"
        )
        run_parser.add_argument(
            "--not_check_file",
            default=False,
            action="store_true",
            required=False,
            help="use to not check if request_input can be a file or not",
        )
        run_parser.set_defaults(func=run_command_factory)

    def __init__(
        self,
        id: int = None,
        auth_file: str = None,
        scope: str = "openid",
        domain: str = "live",
        not_use_cache: bool = False,
        cache_dir: str = "~/.cache/elg",
        request_input: str = None,
        request_type: str = "type",
        sync_mode: bool = False,
        timeout: int = None,
        not_check_file: bool = True,
    ):
        self._id = id
        self._auth_file = auth_file
        self._scope = scope
        self._domain = domain
        self._use_cache = not not_use_cache
        self._cache_dir = cache_dir
        self._request_input = request_input
        self._request_type = request_type
        self._sync_mode = sync_mode
        self._timeout = timeout
        self._check_file = not not_check_file

    def run(self):
        from pprint import pformat

        from ..service import Service

        try:
            service = Service.from_id(
                id=self._id,
                auth_file=self._auth_file,
                scope=self._scope,
                domain=self._domain,
                use_cache=self._use_cache,
                cache_dir=self._cache_dir,
            )
            logger.info(f"Calling {service.resource_name}")
            result = service(
                request_input=self._request_input,
                request_type=self._request_type,
                sync_mode=self._sync_mode,
                timeout=self._timeout,
                check_file=self._check_file,
                verbose=True,
            )
            print(f"Result:\n\n{pformat(result.dict(), indent=2)}\n")
        except Exception as e:
            logger.error(f"Error during service call - {e}")
            sys.exit(1)



import concurrent.futures as cf
import time
from typing import Callable, List, Union

import pandas as pd
from tqdm import tqdm

from .authentication import Authentication
from .entity import Entity
from .model import Request
from .service import Service


class Benchmark:
    """
    Class to execute multiple services in parallel and allows for easy comparison of their outputs using the same input.

    Examples::

        from elg import Benchmark

        # A benchmark can be initialized with a list of services ids (from_ids method) or with a list of entities (from_entities method).
        # Here we compare English to German Machine Translation services.
        ben = Benchmark.from_ids([610, 624])

        # The benchmark can be run on multiple inputs and can be run multiple times to guarantee the result (the first run is also
        # usually longer than the next ones due to the service pods initialization).
        result = ben(["Bush is the president of the USA and lives in Washington.", "ELG is an amazing project."], number_of_runs=2)

        # The benchmark call returns a benchmark result object that can be used to compare the result.

        # You can have an overview of the result,
        df = result.compare()
        print("General comparison:\\n", df)

        # compare only the results,
        df = result.compare_results()
        print("Comparison of the results:\\n", df)

        # or only the response time.
        df = result.compare_response_times()
        print("Comparison of the response time:\\n", df)

        # The compare methods return a DataFrame object that can be exported to csv, excel and many other formats to have a
        # better visualization
        result.compare().to_csv("/tmp/result.csv")

        # We can take another example and compare sentiment analysis services.
        ben = Benchmark.from_ids([477, 510])
        inputs = [
            "This movie is not good at all.",
            "This movie is not good but it was a good moment at the cinema.",
            "This movie is not so bad.",
            "I liked the movie but it was not must seen.",
            "It was the best movie I have ever seen."
        ]
        result = ben(
            inputs,
            output_funcs=[
                lambda x: x.features["OVERALL"],
                lambda x: x.annotations["SentenceSet"][0].features["score"] * 100
            ]
        )

        print("Result:\\n")
        result.compare()
    """

    def __init__(self, services: List[Service]):
        """
        Init a Benchmark with the services to compare

        Args:
            services (List[Service]): list of the services to compare.
        """
        for service in services:
            assert isinstance(service, Service)
        self.services = services

    @classmethod
    def from_entities(
        cls,
        entities: List[Entity],
        auth_object: str = None,
        auth_file: str = None,
        scope: str = None,
        use_cache: bool = True,
        cache_dir="~/.cache/elg",
    ):
        """
        Class method to init a Benchmark using a list of entities which will be convert into services
        using the `from_entity` class method of the Service class. Refer to this method for further explanation.
        """
        return cls(
            [
                Service.from_entity(
                    entity=entity,
                    auth_object=auth_object,
                    auth_file=auth_file,
                    scope=scope,
                    use_cache=use_cache,
                    cache_dir=cache_dir,
                )
                for entity in entities
            ]
        )

    @classmethod
    def from_ids(
        cls,
        ids: List[int],
        auth_object: Authentication = None,
        auth_file: str = None,
        scope: str = None,
        domain: str = None,
        use_cache: bool = True,
        cache_dir: str = "~/.cache/elg",
    ):
        """
        Class method to init a Benchmark using a list of ids which will be convert into services
        using the `from_id` class method of the Service class. Refer to this method for further explanation.
        """
        return cls(
            [
                Service.from_id(
                    id=id,
                    auth_object=auth_object,
                    auth_file=auth_file,
                    scope=scope,
                    domain=domain,
                    use_cache=use_cache,
                    cache_dir=cache_dir,
                )
                for id in ids
            ]
        )

    def __call__(
        self,
        request_inputs: Union[str, List[str], Request, List[Request]] = None,
        request_type: str = "text",
        sync_mode: bool = False,
        timeout: int = None,
        check_file: bool = True,
        output_funcs: Union[str, Callable, List[Union[str, Callable]]] = "auto",
        number_of_runs: int = 2,
    ):
        """
        Method to run the comparison of the services with the given inputs.

        Args:
            request_inputs (Union[str, List[str], Request, List[Request]], optional): list of inputs on which to compare the services.
                Each input must correspond to the request_input parameter of the Service __call__ method. Defaults to None.
            request_type (str, optional): precise the type of the request. Can be "text", "structuredText", or "audio".
                It is only used if request_input is not a Request object. Defaults to "text".
            sync_mode (bool, optional): sync_mode parameter to give to the Service __call__ method. Defaults to False.
            timeout (int, optional): timeout parameter to give to the Service __call__ method. Defaults to None.
            check_file (bool, optional): check_file parameter to give to the Service __call__ method. Defaults to True.
            output_funcs (Union[str, Callable, List[Union[str, Callable]]], optional): output_func parameters to give to the Services
                __call__ method. Defaults to "auto".
            number_of_runs (int, optional): number of times to run the services on each input. It is recommended to run the services
                at least 2 times because on the first time the services usually need to be loaded in the ELG cluster, which
                will increase the response time. The response time of the second pass is, therefore, more precise. Defaults to 2.

        Returns:
            BenchmarkResult: result of the Benchmark call. To obtain the pandas DataFrame containing all the results, run the compare
                method on the obtained BenchmarkResult object.
        """
        if not isinstance(request_inputs, List):
            request_inputs = [request_inputs]

        if isinstance(output_funcs, str) and output_funcs != "auto":
            raise ValueError(
                "output_funcs must be set to 'auto', a callable object, or a list of 'auto' or callable objects with the same length as the number of services in self.services."
            )
        elif isinstance(output_funcs, str) or isinstance(output_funcs, Callable):
            output_funcs = [output_funcs for _ in range(len(self.services))]
        elif isinstance(output_funcs, List):
            if len(output_funcs) != len(self.services):
                raise ValueError(
                    "output_funcs must be set to 'auto', a callable object, or a list of 'auto' or callable objects with the same length as the number of services in self.services."
                )
            for output_func in output_funcs:
                if not ((isinstance(output_func, str) and output_func == "auto") or isinstance(output_func, Callable)):
                    raise ValueError(
                        "output_funcs must be set to 'auto', a callable object, or a list of 'auto' or callable objects with the same length as the number of services in self.services."
                    )
        else:
            raise ValueError(
                "output_funcs must be set to 'auto', a callable object, or a list of 'auto' or callable objects with the same length as the number of services in self.services."
            )

        benchmark_result = BenchmarkResult(services=self.services, request_inputs=request_inputs)

        for request_input in tqdm(request_inputs, desc="Input strs"):
            for run_idx in tqdm(range(number_of_runs), desc="Number of runs"):
                with cf.ThreadPoolExecutor() as executor:
                    futures = []
                    for output_func, service in zip(output_funcs, self.services):
                        futures.append(
                            executor.submit(
                                self._call_service,
                                service=service,
                                request_input=request_input,
                                request_type=request_type,
                                sync_mode=sync_mode,
                                timeout=timeout,
                                check_file=check_file,
                                output_func=output_func,
                            )
                        )

                    for future in cf.as_completed(futures):
                        service, result, response_time = future.result()
                        benchmark_result._append_run(
                            service=service,
                            request_input=request_input,
                            run_idx=run_idx,
                            result=result,
                            response_time=response_time,
                        )

        benchmark_result._set_df()
        return benchmark_result

    @staticmethod
    def _call_service(service: Service, **kwargs) -> (dict, float):
        """
        Internal method to call a service
        """
        start = time.time()
        try:
            result = service(verbose=False, **kwargs)
            return service, result, time.time() - start
        except Exception as e:
            return service, {"response": str(e)}, 0.0


class BenchmarkResult:
    """
    Class the represent the result of a Benchmark call
    """

    def __init__(self, services: List[Service], request_inputs: List[str]):
        """
        Init method that taked the list of the services compared and the inputs on which to compare the services
        """
        self.services = services
        self.id2service = {service.id: service for service in self.services}
        self.result = {service: {str(request_input): {} for request_input in request_inputs} for service in services}

    def set_colwidth(self, value: int = None):
        """
        Method to easily change the colwidth value of pandas to better vizualize the DataFrame

        Args:
            value (int, optional): value of the colwidth. Defaults to None.
        """
        pd.set_option("max_colwidth", value)

    def _append_run(self, service: Service, request_input: str, run_idx: int, result: str, response_time: float):
        """
        Internal method to append the result of a run. Used by the Benchmark call function
        """
        self.result[service][request_input][run_idx] = {
            "result": result,
            "response_time": response_time,
        }

    def _set_df(self):
        """
        Internal method to create the DataFrame when the Benchmark call is finished
        """
        self.df = pd.DataFrame.from_dict(
            {
                (service.id, str(request_input), run_idx): self.result[service][str(request_input)][run_idx]
                for service in self.result.keys()
                for request_input in self.result[service].keys()
                for run_idx in self.result[service][str(request_input)].keys()
            },
            orient="index",
        )
        self.df.index.set_names(["service", "request_input", "run"], inplace=True)

    def compare(
        self,
        columns: List[str] = ["result", "response_time"],
        func: Union[str, list, dict] = "last",
        level: str = "run",
        colwidth: int = 0,
        **agg_kwargs,
    ):
        """
        Method to compare the obtained results. It returns a pandas DataFrame object containing the comparison

        Args:
            columns (List[str], optional): colums of the DataFrame to returned. Defaults to ["result", "response_time"].
            func (Union[str, list, dict], optional): function to use for the comparison. To see all the possible function,
                please see https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.core.groupby.DataFrameGroupBy.aggregate.html?.
                Defaults to "last".
            level (str, optional): level of the comparison. The level value can be: 'service', 'request_input
            ', or 'run'. Defaults to "run".
            colwidth (int, optional): if set, will change the colwidth parameter of pandas to better vizualize the DataFrame. Defaults to 0.

        Raises:
            ValueError: error if the level parameters is not set to a correct value.

        Returns:
            pd.DataFrame: pandas DataFrame object containing the comparison
        """
        if (isinstance(colwidth, int) and colwidth > 0) or colwidth is None:
            self.set_colwidth(colwidth)

        if level == "service":
            level = ["service"]
        elif level == "request_input":
            level = ["service", "request_input"]
        elif level == "run":
            level = ["service", "request_input", "run"]
        else:
            raise ValueError("level must be set to 'service', 'request_input', or 'run'.")

        return self.df.groupby(level).agg(func=func, **agg_kwargs)[columns]

    def compare_results(
        self,
        func: Union[str, list, dict] = "last",
        level: str = "request_input",
        colwidth: int = 0,
        **agg_kwargs,
    ):
        """
        Method similar to the compare method with default parameters optimized to compare the results
        """
        return self.compare(
            columns=["result"],
            func=func,
            level=level,
            colwidth=colwidth,
            **agg_kwargs,
        )

    def compare_response_times(
        self,
        func: Union[str, list, dict] = "describe",
        level: str = "service",
        colwidth: int = 0,
        **agg_kwargs,
    ):
        """
        Method similar to the compare method with default parameters optimized to compare the response_times
        """
        return self.compare(
            columns=["response_time"],
            func=func,
            level=level,
            colwidth=colwidth,
            **agg_kwargs,
        )



from typing import Dict, List

from pydantic import BaseModel, root_validator

from .. import Annotation, ResponseObject


class TextsResponseObject(BaseModel):
    """
    Object representing a structured piece of text. Recursive.

    Attributes
    ----------
    role: (str, optional) the role of this node in the response
    content:(str, optional) string of translated/transcribed text
    texts: (List[TextsResponseObject], optional) list of same structures, recursive
    score: (int, optional) confidence of response
    features (dict, optional): arbitrary JSON metadata about content
    annotations (Dict[str, List[Annotation]], optional): optional annotations on request
    """

    role: str = None
    content: str = None
    texts: List = None
    score: int = None
    features: dict = None
    annotations: Dict[str, List[Annotation]] = None

    @root_validator()
    def either_content_or_text(cls, values):
        """ "
        validator: ensures either the "content" or "text" fields are present
        """
        content, texts, score = values.get("content"), values.get("texts"), values.get("score")
        if content is None and texts is None:
            raise ValueError('A text response must have either "content" or "texts" fields')
        if texts and score:
            raise ValueError("A branch node should not have a score attached")
        return values

    def auto_content(self):
        if self.features is not None and self.annotations is not None:
            return self.dict()
        if self.content is not None and self.texts is None:
            if self.score is None:
                return self.content
            return {
                "content": self.content,
                "score": self.score,
            }
        if self.content is None and self.texts is not None:
            if len(self.texts) == 1:
                return self.texts[0]
            return self.texts


class TextsResponse(ResponseObject):
    """
    Response consisting of a set of one or more new texts, each with optional annotations attached to it.

    For example a set of possible translations produced by a translation tool or possible transcriptions produced by a
    speech-to-text recogniser.

    Attributes
    ----------
    type (str, required): type of response
    warnings (List[StatusMessage], optional): messages describing any warnings on responses
    texts (List[TextsResponseObject], optional): list of objects representing a structured text response
    """

    type: str = "texts"
    texts: List[TextsResponseObject]

    def auto_content(self):
        if self.warnings is not None:
            return self.dict()
        if len(self.texts) == 1:
            return self.texts[0].auto_content()
        return self.dict()["texts"]



import inspect
import json
import re
from pathlib import Path
from posixpath import expanduser
from typing import Dict, List

from loguru import logger

try:
    import docker
    from flask import Flask
    from flask import request as flask_request
    from flask_json import JsonError, as_json
    from requests_toolbelt import MultipartDecoder
except:
    raise ImportError(
        "Extra dependencies need to be install to use the FlaskService class. Please run: `pip install elg[provider]`."
    )

from .model import AudioRequest, StructuredTextRequest, TextRequest
from .utils.docker import COPY_FOLDER, DOCKERFILE, ENTRYPOINT


class FlaskService:
    """
    Class to help the creation of an ELG compatible service from a python model.
    Extra dependencies need to be install to use the FlaskService class. Please run: `pip install elg[provider]`.
    """

    requirements = ["Flask", "Flask-JSON", "gunicorn", "elg", "docker", "flask", "flask_json"]

    def __init__(self, name: str):
        """
        Init function of the FlaskService

        Args:
            name (str): Name of the service. It doesn't have any importance.
        """
        self.name = name
        self.app = Flask(name)
        # Don't add an extra "status" property to JSON responses - this would break the API contract
        self.app.config["JSON_ADD_STATUS"] = False
        # Don't sort properties alphabetically in response JSON
        self.app.config["JSON_SORT_KEYS"] = False
        self.app.add_url_rule("/process", "process", self.process, methods=["POST"])

    def run(self):
        """
        Method to start the flask app.
        """
        self.app.run()

    @as_json
    def process(self):
        """
        Main request processing logic - accepts a JSON request and returns a JSON response.
        """
        logger.info("Process request")
        if "application/json" in flask_request.content_type:
            data = flask_request.get_json()
        elif "multipart/form-data" in flask_request.content_type:
            decoder = MultipartDecoder(flask_request.get_data(), flask_request.content_type)
            data = {}
            for part in decoder.parts:
                headers = {k.decode(): v.decode() for k, v in part.headers.items()}
                if "application/json" in headers["Content-Type"]:
                    for k, v in json.loads(part.content.decode()).items():
                        data[k] = v
                elif "audio" in headers["Content-Type"]:
                    data["content"] = part.content
                else:
                    raise ValueError("Unknown Content-Type in multipart request")
        else:
            raise ValueError()

        if data.get("type") == "audio":
            request = AudioRequest(**data)
        elif data.get("type") == "text":
            request = TextRequest(**data)
        elif data.get("type") == "structuredText":
            request = StructuredTextRequest(**data)
        else:
            self.invalid_request_error()
        logger.info(f"Call with the input: {request}")
        response = self.process_request(request)
        logger.info(f"Get response: {response}")
        response = {"response": response.dict(by_alias=True)}
        logger.info(f"Return: {response}")
        return response

    def invalid_request_error(self):
        """
        Generates a valid ELG "failure" response if the request cannot be parsed
        """
        raise JsonError(
            status_=400, failure={"errors": [{"code": "elg.request.invalid", "text": "Invalid request message"}]}
        )

    def process_request(self, request):
        """
        Method to process the request object. This method only calls the right process method regarding the type of the request.
        """
        if request.type == "text":
            return self.process_text(request)
        elif request.type == "structuredText":
            return self.process_structured_text(request)
        elif request.type == "audio":
            return self.process_audio(request)
        self.invalid_request_error()

    def process_text(self, request: TextRequest):
        """
        Method to implement if the service takes text as input.

        Args:
            request (TextRequest): TextRequest object.
        """
        raise NotImplementedError()

    def process_structured_text(self, request: StructuredTextRequest):
        """
        Method to implement if the service takes structured text as input.

        Args:
            request (StructuredTextRequest): StructuredTextRequest object.
        """
        raise NotImplementedError()

    def process_audio(self, request: AudioRequest):
        """
        Method to implement if the service takes audio as input.

        Args:
            request (AudioRequest): AudioRequest object.
        """
        raise NotImplementedError()

    @classmethod
    def create_requirements(cls, requirements: List = [], path: str = None):
        """
        Class method to create the correct requirements.txt file.

        Args:
            requirements (List, optional): List of required pip packages. Defaults to [].
            path (str, optional): Path where to generate the file. Defaults to None.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        requirements = cls.requirements + requirements
        with open(path.parent / "requirements.txt", "w") as f:
            f.write("\n".join(set(requirements)))

    @classmethod
    def create_docker_files(cls, required_files: List = [], required_folders: List = [], path: str = None):
        """Class method to create the correct Dockerfile.

        Args:
            required_files (List, optional): List of files needed for the service. Defaults to [].
            required_folders (List, optional): List of folders needed for the service. Defaults to [].
            path (str, optional): Path where to generate the file. Defaults to None.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        required_files = [path.name] + required_files
        required_folders = "\n".join(
            [COPY_FOLDER.format(folder_name=str(Path(folder))) for folder in required_folders]
        )
        service_script = path.name[:-3]  # to remove .py
        # The docker-entrypoint file is a Linux shell script so _must_ be
        # written with Unix-style line endings, even if the build is being done
        # on Windows.  To ensure this we write in binary mode.
        with open(path.parent / "docker-entrypoint.sh", "wb") as f:
            f.write(ENTRYPOINT.format(service_script=service_script).encode('utf-8'))
        with open(path.parent / "Dockerfile", "w") as f:
            f.write(DOCKERFILE.format(required_files=" ".join(required_files), required_folders=required_folders))

    @classmethod
    def docker_build_image(cls, tag: str, pull: bool = True, path: str = None, **kwargs):
        """
        Class method to do `docker build ...` in python.
        """
        if path == None:
            path = Path(inspect.getsourcefile(cls))
        else:
            path = Path(path)
        client = docker.from_env()
        image, _ = client.images.build(path=str(path.parent), tag=tag, pull=pull, **kwargs)
        return image

    @classmethod
    def docker_push_image(cls, repository: str, tag: str, username: str = None, password: str = None, **kwargs):
        """
        Class method to do `docker push ...` in python.
        """
        client = docker.from_env()
        if username is not None and password is not None:
            auth_config = {"username": username, "password": password}
            client.images.push(repository=repository, tag=tag, auth_config=auth_config, stream=True, **kwargs)
        client.images.push(repository=repository, tag=tag, stream=True, **kwargs)
        return

    @classmethod
    def docker_build_push_image(
        cls,
        repository: str,
        tag: str,
        pull: bool = True,
        username: str = None,
        password: str = None,
        build_kwargs: Dict = {},
        push_kwargs: Dict = {},
    ):
        cls.docker_build_image(tag=f"{repository}:{tag}", pull=pull, **build_kwargs)
        cls.docker_push_image(repository=repository, tag=tag, username=username, password=password, **push_kwargs)
        return None



import sys
from argparse import ArgumentParser

from loguru import logger

from . import BaseELGCommand


def download_command_factory(args):
    return DownloadCommand(
        id=args.id,
        domain=args.domain,
        not_use_cache=args.not_use_cache,
        cache_dir=args.cache_dir,
        distribution_idx=args.distribution_idx,
        filename=args.filename,
        folder=args.folder,
    )


class DownloadCommand(BaseELGCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        download_parser = parser.add_parser("download", description="Download an ELG corpus")
        download_parser.add_argument(
            "--id", type=int, default=None, required=True, help="ID of the corpus to download"
        )
        download_parser.add_argument(
            "--domain",
            type=str,
            default="live",
            required=False,
            help="ELG domain you want to use",
        )
        download_parser.add_argument(
            "--not_use_cache",
            default=False,
            action="store_true",
            required=False,
            help="use to not use cache",
        )
        download_parser.add_argument(
            "--cache_dir", type=str, default="~/.cache/elg", required=False, help="path to the cache directory"
        )
        download_parser.add_argument(
            "--distribution_idx",
            type=int,
            default=0,
            required=False,
            help="index of the distribution of the corpus to download",
        )
        download_parser.add_argument(
            "--filename",
            default=None,
            type=str,
            required=False,
            help="name of the output file. If None, the name of the corpus will be used",
        )
        download_parser.add_argument(
            "--folder",
            type=str,
            default="./",
            required=False,
            help="path to the folder where to save the downloaded file",
        )
        download_parser.set_defaults(func=download_command_factory)

    def __init__(
        self,
        id: int = None,
        domain: str = "live",
        not_use_cache: bool = False,
        cache_dir: str = "~/.cache/elg",
        distribution_idx: int = 0,
        filename: str = None,
        folder: str = "./",
    ):
        self._id = id
        self._domain = domain
        self._use_cache = not not_use_cache
        self._cache_dir = cache_dir
        self._distribution_idx = distribution_idx
        self._filename = filename
        self._folder = folder

    def run(self):
        from ..corpus import Corpus

        try:
            corpus = Corpus.from_id(
                id=self._id,
                domain=self._domain,
                use_cache=self._use_cache,
                cache_dir=self._cache_dir,
            )
            logger.info(f"Downloading {corpus.resource_name}")
            corpus.download(
                distribution_idx=self._distribution_idx,
                filename=self._filename,
                folder=self._folder,
            )
        except Exception as e:
            logger.error(f"Error during corpus download - {e}")
            sys.exit(1)



from pathlib import Path
from typing import List

import requests
from tqdm import tqdm

from .authentication import (Authentication, NeedAuthentication,
                             need_authentication)
from .entity import Entity
from .utils import (CORPUS_DOWNLOAD_URL, LICENCE_URL, get_domain, get_en_value,
                    get_information, get_metadatarecord,
                    map_metadatarecord_to_result)
from .utils.errors import (NotCorpusException, catch_requests_error,
                           ensure_response_ok)


class Licence:
    """
    Class to represent a licence
    """

    def __init__(
        self,
        name: str,
        urls: List[str],
        identifiers: List[dict],
    ):
        self.name = name
        self.urls = urls
        self.identifiers = identifiers

    @classmethod
    def from_data(cls, data: dict):
        name = get_en_value(get_information(id=-1, obj=data, infos="licence_terms_name"))
        urls = get_information(id=-1, obj=data, infos="licence_terms_url")
        identifiers = get_information(id=-1, obj=data, infos="licence_identifier")
        return cls(
            name=name,
            urls=urls,
            identifiers=identifiers,
        )

    def _get_elg_licence_identifier(self):
        for identifier in self.identifiers:
            if identifier["licence_identifier_scheme"] == "http://w3id.org/meta-share/meta-share/elg":
                return identifier["value"]
        return None


class Distribution:
    """
    Class to represent a corpus distribution
    """

    def __init__(
        self,
        corpus_id: int,
        domain: str,
        form: str,
        distribution_location: str,
        download_location: str,
        access_location: str,
        licence: Licence,
        cost: str,
        attribution_text: str,
    ):
        self.corpus_id = corpus_id
        self.domain = domain
        self.form = form
        self.distribution_location = distribution_location
        self.download_location = download_location
        self.access_location = access_location
        self.licence = licence
        self.cost = cost
        self.attribution_text = attribution_text

    @classmethod
    def from_data(cls, corpus_id: int, domain: str, data: dict):
        """
        Class method to init the distribution object from the metadata information.

        Args:
            corpus_id (int): id of the corpus the distribution is from.
            domain (str): ELG domain you want to use. "live" to use the public ELG, "dev" to use the development ELG and
                another value to use a local ELG.
            data (dict): metadata information of the distribution.

        Returns:
            elg.Distribution: the distribution object initialized.
        """
        form = get_information(id=-1, obj=data, infos="dataset_distribution_form")
        distribution_location = get_information(id=-1, obj=data, infos="distribution_location", none=True)
        download_location = get_information(id=-1, obj=data, infos="download_location", none=True)
        access_location = get_information(id=-1, obj=data, infos="access_location", none=True)
        licence_terms = get_information(id=-1, obj=data, infos="licence_terms")
        assert isinstance(licence_terms, list) and len(licence_terms) == 1, licence_terms
        licence = Licence.from_data(licence_terms[0])
        cost = get_information(id=-1, obj=data, infos="cost", none=True)
        if cost is not None:
            cost = "{} {}".format(
                get_information(id=-1, obj=cost, infos="amount"),
                get_information(id=-1, obj=cost, infos="currency").split("/")[-1],
            )
        attribution_text = get_information(id=-1, obj=data, infos="attribution_text", none=True)
        if attribution_text is not None:
            attribution_text = get_en_value(attribution_text)
        return cls(
            corpus_id=corpus_id,
            domain=domain,
            form=form,
            distribution_location=distribution_location,
            download_location=download_location,
            access_location=access_location,
            licence=licence,
            cost=cost,
            attribution_text=attribution_text,
        )

    def is_downloadable(self) -> str:
        """
        Method to get if the distribution is downloadable.

        Returns:
            bool: return True is the distribution is downloadable, False if not.
        """
        if self.form == "http://w3id.org/meta-share/meta-share/downloadable":
            return True
        return False

    @catch_requests_error
    def get_download_url(self, access_token: str) -> str:
        """
        Method to obtain the download url of the distribution if it exists.

        Returns:
            str: url to download the dataset if it exists, else None.
        """
        if self.access_location is not None and self.access_location != "":
            not_downloadable = f"This corpus distribution is not downloadable from the ELG. The corpus is accessible by clicking: {self.access_location}"
        elif self.distribution_location is not None and self.distribution_location != "":
            not_downloadable = f"This corpus distribution is not downloadable from the ELG. The corpus is distributed at the following URL: {self.distribution_location}"
        else:
            not_downloadable = "This corpus distribution is not downloadable from the ELG."
        if not self.is_downloadable():
            print(not_downloadable)
            return None
        if self.download_location is not None and self.download_location != "":
            return self.download_location
        elg_licence_identifier = self.licence._get_elg_licence_identifier()
        if elg_licence_identifier is not None:
            elg_licence_url = LICENCE_URL.format(domain=self.domain, licence=elg_licence_identifier)
            # check if the elg_licence_url exists
            response = requests.get(elg_licence_url)
            if not response.ok:
                print(not_downloadable)
                return None
            print(f"Please, visit the licence of this corpus distribution by clicking: {elg_licence_url}\n")
            accept = input("Do you accept the licence terms: (yes/[no]): ")
            if accept not in ["yes", "Yes", "y", "Y"]:
                print("You need to accept the licence terms to download this corpus distribution.")
                return None
        data = {"licences": [f"{elg_licence_identifier}"]}
        headers = {"Authorization": f"Bearer {access_token}"}
        response = requests.post(
            CORPUS_DOWNLOAD_URL.format(domain=self.domain, id=self.corpus_id), json=data, headers=headers
        )
        if response.ok:
            return response.json()["s3-url"]
        print(not_downloadable)
        return None

    @catch_requests_error
    def download(self, access_token: str, filename: str = None, folder: str = "./"):
        """
        Method to download the distribution if possible.

        Args:
            filename (str, optional): Name of the output file. If None, the name of the downloaded file will be used.
                Defaults to None.
            folder (str, optional): path to the folder where to save the downloaded file. Defaults to "./".
        """
        url = self.get_download_url(access_token=access_token)
        if url is None:
            return None
        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)
        if isinstance(filename, str):
            # use the given filename with the correct extension
            extension = url.split("?")[0].split(".")[-1]
            filename = folder / Path(filename + ("." + extension if len(extension) <= 5 else ""))
        else:
            # use the filename of the file to download
            filename = folder / Path(url.split("?")[0].split("/")[-1])
        print(f"\nDownloading the corpus distribution to {filename}:")
        response = requests.get(url, stream=True)
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        block_size = 1024
        progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
        with open(filename, "wb") as f:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                f.write(data)
        progress_bar.close()
        ensure_response_ok(response)
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            print("Something went wrong during the downloading. The file may be not usable.")


class Corpus(Entity, NeedAuthentication):
    """
    Class to represent a corpus. Download ELG corpora.

    Examples::

        from elg import Corpus

        # You can initialize a corpus from its id. You will be asked to authenticate on the ELG website.
        corpus = Corpus.from_id(913)

        # You can display the corpus information.
        print(corpus)

        # You can download the corpus. Note that only corpora hosted on ELG are downloadable using the python SDK.
        corpus.download()

        # By default the corpus is downloaded at the current location and the filename is the name of the ELG corpus.
        # You can overwrite this with the folder and filename parameters.
        corpus.download(filename="ELG_corpus", folder="/tmp/")

        # You can create an corpus from a catalog search result. First you need to search for a service using the catalog.
        # Let's search an English to French Machine Translation service.
        from elg import Catalog

        catalog = Catalog()
        results = catalog.search(
            resource = "Corpus",
            languages = ["German"],
            search="ner",
            limit = 1,
        )

        corpus = Corpus.from_entity(results[0])
        print(corpus)

    """

    def __init__(
        self,
        id: int,
        resource_name: str,
        resource_short_name: List[str],
        resource_type: str,
        entity_type: str,
        description: str,
        keywords: List[str],
        detail: str,
        licences: List[str],
        languages: List[str],
        country_of_registration: List[str],
        creation_date: str,
        last_date_updated: str,
        functional_service: bool,
        functions: List[str],
        intended_applications: List[str],
        views: int,
        downloads: int,
        size: int,
        service_execution_count: int,
        status: str,
        under_construction: bool,
        record: dict,
        auth_object: Authentication,
        auth_file: str,
        scope: str,
        domain: str,
        use_cache: bool,
        cache_dir: str,
    ):
        """
        Init a Corpus object with the corpus information
        """

        super().__init__(
            id=id,
            resource_name=resource_name,
            resource_short_name=resource_short_name,
            resource_type=resource_type,
            entity_type=entity_type,
            description=description,
            keywords=keywords,
            detail=detail,
            licences=licences,
            languages=languages,
            country_of_registration=country_of_registration,
            creation_date=creation_date,
            last_date_updated=last_date_updated,
            functional_service=functional_service,
            functions=functions,
            intended_applications=intended_applications,
            views=views,
            downloads=downloads,
            size=size,
            service_execution_count=service_execution_count,
            status=status,
            under_construction=under_construction,
            domain=domain,
            record=record,
        )
        if self.resource_type != "Corpus":
            raise NotCorpusException(self.id)
        self._authenticate(
            auth_object=auth_object, auth_file=auth_file, scope=scope, use_cache=use_cache, cache_dir=cache_dir
        )
        self.distributions = [
            Distribution.from_data(self.id, self.domain, data)
            for data in get_information(
                id=id, obj=self.record, infos=["described_entity", "lr_subclass", "dataset_distribution"]
            )
        ]

    @classmethod
    def from_id(
        cls,
        id: int,
        auth_object: Authentication = None,
        auth_file: str = None,
        scope: str = None,
        domain: str = None,
        use_cache: bool = True,
        cache_dir: str = "~/.cache/elg",
    ):
        """
        Class method to init a Corpus class from its id. You can provide authentication information through the
        auth_object or the auth_file attributes. If not authentication information is provided, the Authentication
        object will be initialized.

        Args:
            id (int): id of the corpus.
            auth_object (elg.Authentication, optional): elg.Authentication object to use. Defaults to None.
            auth_file (str, optional): json file that contains the authentication tokens. Defaults to None.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".
            use_cache (bool, optional): True if you want to use cached files. Defaults to True.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files. Defaults to "~/.cache/elg".

        Returns:
            elg.Corpus: the corpus object initialized.
        """
        # Use auth_object/file domain if auth_object/file given
        if auth_object is not None:
            assert isinstance(
                auth_object, Authentication
            ), f"auth_object ({type(auth_object)}) must be an Authentication object."
            auth_domain = auth_object.domain
            if domain is not None:
                print(
                    "Warning: domain argument saved in the authentication object will overwrite the "
                    "domain argument you put."
                )
            domain = auth_domain
        elif isinstance(auth_file, str):
            auth_domain = get_argument_from_json(auth_file, "domain")
            if domain is not None:
                print(
                    "Warning: domain argument saved in the authentication filename will overwrite the "
                    "domain argument you put."
                )
            domain = auth_domain
        # Use "live" domain by default
        domain = get_domain(domain=domain if domain is not None else "live")
        record = get_metadatarecord(id=id, domain=domain, use_cache=use_cache, cache_dir=cache_dir)
        result = map_metadatarecord_to_result(id, record)
        result["auth_object"] = auth_object
        result["auth_file"] = auth_file
        result["scope"] = scope
        result["domain"] = domain
        result["use_cache"] = use_cache
        result["cache_dir"] = cache_dir
        return cls(**result)

    @classmethod
    def from_entity(
        cls,
        entity: Entity,
        auth_object: str = None,
        auth_file: str = None,
        scope: str = None,
        use_cache: bool = True,
        cache_dir="~/.cache/elg",
    ):
        """
        Class method to init a Corpus class from an Entity object. You can provide authentication information through the
        auth_object or the auth_file attributes. If not authentication information is provided, the Authentication object will be initialized.

        Args:
            entity (elg.Entity): Entity object to init as a Corpus.
            auth_object (elg.Authentication, optional): elg.Authentication object to use. Defaults to None.
            auth_file (str, optional): json file that contains the authentication tokens. Defaults to None.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".
            use_cache (bool, optional): True if you want to use cached files. Defaults to True.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files. Defaults to "~/.cache/elg".

        Returns:
            elg.Corpus: Corpus object with authentication information.
        """
        if entity.record is None:
            entity.record = get_metadatarecord(
                id=entity.id, domain=entity.domain, use_cache=use_cache, cache_dir=cache_dir
            )
        parameters = entity.__dict__
        parameters["auth_object"] = auth_object
        parameters["auth_file"] = auth_file
        parameters["scope"] = scope
        parameters["use_cache"] = use_cache
        parameters["cache_dir"] = cache_dir
        return cls(**parameters)

    @need_authentication()
    def download(self, distribution_idx: int = 0, filename: str = None, folder: str = "./"):
        """
        Method to download the corpus if possible.

        Args:
            distribution_idx (int, optional): Index of the distribution of the corpus to download. Defaults to 0.
            filename (str, optional): Name of the output file. If None, the name of the corpus will be used. Defaults to None.
            folder (str, optional): path to the folder where to save the downloaded file. Defaults to "./".
        """
        print(f"Downloading:\n\t[{self.id}] {self.resource_name}\n")
        filename = (
            filename
            if filename is not None
            else "_".join(
                [w for w in "".join(c if c.isalnum() else " " for c in self.resource_name).split(" ") if w != ""]
            )
        )
        self.distributions[distribution_idx].download(
            access_token=self.authentication.access_token,
            filename=filename,
            folder=folder,
        )
